package jp.topse.kaizu.sequence2csp.convert;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.BehaviorExecutionSpecification;
import org.eclipse.uml2.uml.Interaction;
import org.eclipse.uml2.uml.InteractionFragment;
import org.eclipse.uml2.uml.Lifeline;
import org.eclipse.uml2.uml.Message;
import org.eclipse.uml2.uml.MessageEnd;
import org.eclipse.uml2.uml.MessageOccurrenceSpecification;
import org.eclipse.uml2.uml.OccurrenceSpecification;
import org.eclipse.uml2.uml.StateInvariant;

public class Converter {
	public static final String CALL_KEYWORD = "call_";

	public static final String COMPO_KEYWORD = "COMPO_";

	public static final String DEFAULT_STATE_KEYWORD = "default_";

	public Converter() {
	}

	/**
	 * ̃V[PX}1FDRLqɕϊ
	 * 
	 * @param interactions
	 * @return
	 * @throws IOException
	 */
	public String convert(List<Interaction> interactions) throws IOException {
		BufferedReader br = new BufferedReader(new InputStreamReader(
				Converter.class.getResourceAsStream("/template.txt")));
		try {
			Set<Lifeline> lifelines = new HashSet<Lifeline>();
			for (Interaction interaction : interactions) {
				for (Object lifeline : interaction.getLifelines()) {
					lifelines.add((Lifeline) lifeline);
				}
			}

			Set<String> lifelineNames = new HashSet<String>();
			Set<String> actorNameSet = new HashSet<String>();
			for (Lifeline lifeline : lifelines) {
				String name = CSPNames.get(lifeline);
				lifelineNames.add(name);
				if (Util.isActor(lifeline)) {
					actorNameSet.add(name);
				}
			}

			Set<String> messageNames = new HashSet<String>();
			for (Interaction interaction : interactions) {
				for (Object message : interaction.getMessages()) {
					messageNames.add(CSPNames.get((Message) message));
				}
			}

			String compoNames = Util.join(lifelineNames, " | ");
			String actorNames = Util.join(actorNameSet, " | ");
			String stateNames = Util.join(DEFAULT_STATE_KEYWORD,
					getLifelineStates(interactions), " | ");
			String callNames = Util.join(messageNames, " | ");
			String compoDef = convert(lifelines);

			StringBuilder sb = new StringBuilder();
			String line;
			while ((line = br.readLine()) != null) {
				line = line.replace("${CompoNames}", compoNames);
				line = line.replace("${ActorNames}", actorNames);
				line = line.replace("${StateNames}", stateNames);
				line = line.replace("${CallNames}", callNames);
				line = line.replace("${CompoDef}", compoDef);
				line = line.replace("${Spec}", "");
				sb.append(line);
				sb.append(Util.LINE_SEPARATOR);
			}

			return sb.toString();
		} finally {
			br.close();
		}
	}

	/**
	 * CtCɑ݂Ԗ̈ꗗԂ
	 * 
	 * @param interactions
	 * @return
	 */
	private Set<String> getLifelineStates(List<Interaction> interactions) {
		Set<String> states = new HashSet<String>();
		for (Interaction interaction : interactions) {
			for (Object fragment : interaction.getFragments()) {
				if (fragment instanceof StateInvariant) {
					StateInvariant si = (StateInvariant) fragment;
					String name = CSPNames.get(si);
					if (!states.contains(name)) {
						states.add(name);
					}
				}
			}
		}
		return states;
	}

	/**
	 * ̃V[PX}1CSPɕϊiCOMPO_̂݁j
	 * 
	 * @param lifelines
	 * @return
	 */
	private String convert(Set<Lifeline> lifelines) {
		Map<String, SequenceTree> forest = new HashMap<String, SequenceTree>();
		for (Lifeline lifeline : lifelines) {
			convert(forest, lifeline);
		}
		StringBuilder sb = new StringBuilder();
		for (Entry<String, SequenceTree> entry : forest.entrySet()) {
			sb.append(COMPO_KEYWORD);
			sb.append("(");
			sb.append(entry.getKey());
			sb.append(") =");
			sb.append(Util.LINE_SEPARATOR);
			sb.append("\t");
			sb.append(entry.getValue().toString());
			sb.append(Util.LINE_SEPARATOR);
			sb.append(Util.LINE_SEPARATOR);
		}
		return sb.toString();
	}

	/**
	 * w肵R|[lgEԂn܂c[Ԃ
	 * 
	 * @param forest
	 * @param compoName
	 * @param startState
	 * @return
	 */
	private SequenceTree getTree(Map<String, SequenceTree> forest,
			String compoName, String startState) {
		String key = compoName + "." + startState;
		if (!forest.containsKey(key)) {
			forest.put(key, SequenceTree.create());
		}
		return forest.get(key);
	}

	/**
	 * 1{̃CtCɒڂăc[ɕϊ
	 * 
	 * @param forest
	 * @param lifeline
	 */
	private void convert(Map<String, SequenceTree> forest, Lifeline lifeline) {
		Interaction interaction = lifeline.getInteraction();
		String compoName = CSPNames.get(lifeline);
		SequenceTree tree = getTree(forest, compoName, DEFAULT_STATE_KEYWORD);
		Map<OccurrenceSpecification, BehaviorExecutionSpecification> besMap = getBesMap(interaction);
		EList list = interaction.getFragments();

		for (int i = 0; i < list.size(); i++) {
			InteractionFragment fragment = (InteractionFragment) list.get(i);
			if (fragment.getCovered(lifeline.getName()) != lifeline) {
				continue;
			}

			if (fragment instanceof MessageOccurrenceSpecification) {
				MessageOccurrenceSpecification mos = (MessageOccurrenceSpecification) fragment;
				// bZ[WM
				tree = convert(tree, mos, lifeline);
				if (besMap.containsKey(mos)) {
					// ԏIc[؂
					InteractionFragment next = getNextFragment(list, i,
							lifeline);
					if (next instanceof StateInvariant) {
						// ̃XebvŒʏStateInvariantƂď
					} else {
						tree.addState(compoName, DEFAULT_STATE_KEYWORD);
						tree = getTree(forest, compoName, DEFAULT_STATE_KEYWORD);
					}
				}

			} else if (fragment instanceof StateInvariant) {
				StateInvariant si = (StateInvariant) fragment;
				// ԕsώ
				String stateName = CSPNames.get(si);
				tree.addState(compoName, stateName);
				tree = getTree(forest, compoName, stateName);
			}
		}
		tree.addState(compoName, DEFAULT_STATE_KEYWORD);
	}

	/**
	 * bZ[WÃbZ[WŏI銈Ԃւ̃}bvB
	 * 
	 * @param interaction
	 * @return
	 */
	private Map<OccurrenceSpecification, BehaviorExecutionSpecification> getBesMap(
			Interaction interaction) {
		Map<OccurrenceSpecification, BehaviorExecutionSpecification> besMap;
		besMap = new HashMap<OccurrenceSpecification, BehaviorExecutionSpecification>();

		for (Object fragment : interaction.getFragments()) {
			if (fragment instanceof BehaviorExecutionSpecification) {
				BehaviorExecutionSpecification bes = (BehaviorExecutionSpecification) fragment;
				besMap.put(bes.getFinish(), bes);
			}
		}

		return besMap;
	}

	/**
	 * ̃CtC̎̃tOg
	 * 
	 * @param list
	 * @param prev
	 * @param lifeline
	 * @return
	 */
	private InteractionFragment getNextFragment(EList list, int prev,
			Lifeline lifeline) {
		for (int i = prev + 1; i < list.size(); i++) {
			InteractionFragment fragment = (InteractionFragment) list.get(i);
			if (fragment.getCovered(lifeline.getName()) == lifeline) {
				return fragment;
			}
		}
		return null;
	}

	/**
	 * 1̃bZ[Wc[ɕϊB
	 * 
	 * @param tree
	 * @param mos
	 * @param lifeline
	 * @return
	 */
	private SequenceTree convert(SequenceTree tree,
			MessageOccurrenceSpecification mos, Lifeline lifeline) {
		Message message = mos.getMessage();

		Lifeline sLifeline = null;
		MessageEnd sendEvent = message.getSendEvent();
		if (sendEvent instanceof InteractionFragment) {
			InteractionFragment sFragment = (InteractionFragment) sendEvent;
			sLifeline = (Lifeline) (sFragment.getCovereds().get(0));
		}

		Lifeline rLifeline = null;
		MessageEnd receiveEvent = message.getReceiveEvent();
		if (receiveEvent instanceof InteractionFragment) {
			InteractionFragment rFragment = (InteractionFragment) receiveEvent;
			rLifeline = (Lifeline) (rFragment.getCovereds().get(0));
		}

		StringBuilder sb = new StringBuilder();
		sb.append(CALL_KEYWORD);
		sb.append(".");
		sb.append(CSPNames.get(message));
		sb.append(".");
		sb.append(CSPNames.get(sLifeline));
		sb.append(".");
		sb.append(CSPNames.get(rLifeline));
		String messageString = sb.toString();
		return tree.addMessage(messageString, (rLifeline == lifeline), Util
				.isActor(lifeline));
	}
}
