package jp.topse.kaizu.sequence2csp.convert;

import java.util.HashMap;
import java.util.Map;

public abstract class SequenceTree {
	public static SequenceTree create() {
		return new SequenceTreeMessage(null);
	}

	public abstract SequenceTree addMessage(String messageString,
			boolean receive, boolean actor);

	public abstract void addState(String compoString, String stateString);

	public abstract String toString(int indent);

	public String toString() {
		return toString(1);
	}

	protected void printIndent(StringBuilder sb, int indent) {
		for (int i = 0; i < indent; i++) {
			sb.append("\t");
		}
	}

	private static class SequenceTreeMessage extends SequenceTree {
		private String messageString;

		private Map<String, SequenceTree> children;

		private boolean canReceive;

		private boolean actorMessage;

		public SequenceTreeMessage(String messageString) {
			this.messageString = messageString;
			this.children = new HashMap<String, SequenceTree>();
			this.canReceive = false;
			this.actorMessage = false;
		}

		@Override
		public SequenceTree addMessage(String message, boolean receive,
				boolean actor) {
			if (receive) {
				canReceive = true;
			}
			if (actor) {
				actorMessage = true;
			}
			if (!children.containsKey(message)) {
				children.put(message, new SequenceTreeMessage(message));
			}
			return children.get(message);
		}

		private boolean isRoot() {
			return (messageString == null);
		}

		@Override
		public void addState(String compoString, String state) {
			if (!isRoot() && !children.containsKey(state)) {
				children.put(state, new SequenceTreeState(compoString, state));
			}
		}

		@Override
		public String toString(int indent) {
			StringBuilder sb = new StringBuilder();
			boolean selection = (children.size() > 1);
			if (messageString != null) {
				sb.append(messageString);
				sb.append(" ->");
				sb.append(Util.LINE_SEPARATOR);
				if (selection) {
					indent++;
				}
				printIndent(sb, indent);
			}
			if (selection) {
				sb.append("(");
			}
			boolean printSelection = false;
			for (SequenceTree child : children.values()) {
				if (printSelection) {
					sb.append(Util.LINE_SEPARATOR);
					printIndent(sb, indent);
					if (canReceive || actorMessage) {
						sb.append("[]");
					} else {
						sb.append("|~|");
					}
					sb.append(Util.LINE_SEPARATOR);
					printIndent(sb, indent);
				}
				sb.append(child.toString(indent));
				printSelection = true;
			}
			if (selection) {
				indent--;
				sb.append(")");
			}
			return sb.toString();
		}
	}

	private static class SequenceTreeState extends SequenceTree {
		public String compoString;

		public String stateString;

		public SequenceTreeState(String compoString, String stateString) {
			this.compoString = compoString;
			this.stateString = stateString;
		}

		@Override
		public SequenceTree addMessage(String message, boolean receive,
				boolean actor) {
			throw new RuntimeException("Internal Error");
		}

		@Override
		public void addState(String compoString, String state) {
			throw new RuntimeException("Internal Error");
		}

		@Override
		public String toString(int indent) {
			return Converter.COMPO_KEYWORD + "(" + compoString + "."
					+ stateString + ")";
		}
	}
}
